@@ -715,8 +715,16 @@ int lttng_event_reserve(struct lttng_ust_lib_ring_buffer_ctx *ctx,
 	}
 
 	ret = lib_ring_buffer_reserve(&client_config, ctx);
-	if (ret)
+	if (caa_unlikely(ret))
 		goto put;
+	if (caa_likely(ctx->ctx_len
+			>= sizeof(struct lttng_ust_lib_ring_buffer_ctx))) {
+		if (lib_ring_buffer_backend_get_pages(&client_config, ctx,
+				&ctx->backend_pages)) {
+			ret = -EPERM;
+			goto put;
+		}
+	}
 	lttng_write_event_header(&client_config, ctx, event_id);
 	return 0;
 put:
